/**
 * ROOT->BLOP and BLOP->ROOT histogram conversions.
 * Only works for uniform binning.
 */
#include "bloproot.h"
#include <TFile.h>
#include <TList.h>
#include <TObject.h>
#include <TClass.h>
#include <TKey.h>
#include <iostream>
#include <iomanip>
using namespace std;


namespace blop{


/// Convert ROOT 1D histogram to BLOP histogram.
void
r2b_convert(const TH1D& r, hist& h)
{
  h.reinit(r.GetBinLowEdge(0),
           r.GetBinLowEdge(r.GetNbinsX()+1)+r.GetBinWidth(r.GetNbinsX()+1),
           r.GetNbinsX()+2,
           r.GetXaxis()->GetTitle(),
           r.GetYaxis()->GetTitle(),
           r.GetTitle());
  for ( unsigned int i=0 ; i<h.size() ; ++i )
  {
    h[i].v=r.GetBinContent(i);
    h[i].e=r.GetBinError(i);
  }
}


/// Convert BLOP 1D histogram to ROOT histogram.
void
b2r_convert(const hist& h, TH1D& r)
{
  r = TH1D(h.legend().c_str(),
           (h.legend()&";"&h.axistitle(1)&";"&h.rangetitle()).c_str(),
           h.bins(1)-2,
           h.min(1)+(h.max(1)-h.min(1))/h.bins(1),
           h.max(1)-(h.max(1)-h.min(1))/h.bins(1));
  for ( unsigned int i=0 ; i<=r.GetNbinsX()+1 ; ++i )
  {
    r.SetBinContent(i, h[i].v);
    r.SetBinError(i, h[i].e);
  }
}


/// Convert ROOT 2D histogram to BLOP histogram.
void
r2b_convert(const TH2D& r, hist& h)
{
  const unsigned int nBinsX = r.GetNbinsX();
  const unsigned int nBinsY = r.GetNbinsY();
  h.reinit(r.GetXaxis()->GetBinLowEdge(0),
           r.GetXaxis()->GetBinLowEdge(nBinsX+1)+r.GetXaxis()->GetBinWidth(nBinsX+1),
           nBinsX+2,
           r.GetXaxis()->GetTitle(),
           r.GetYaxis()->GetBinLowEdge(0),
           r.GetYaxis()->GetBinLowEdge(nBinsY+1)+r.GetYaxis()->GetBinWidth(nBinsY+1),
           nBinsY+2,
           r.GetYaxis()->GetTitle(),
           r.GetZaxis()->GetTitle(),
           r.GetTitle());
  for ( unsigned int binX=0 ; binX<=nBinsX+1 ; ++binX )
  {
    for ( unsigned int binY=0 ; binY<=nBinsY+1 ; ++binY )
    {
      const double binCenterX = r.GetXaxis()->GetBinCenter(binX);
      const double binCenterY = r.GetYaxis()->GetBinCenter(binY);
      const unsigned int i = h.bin(binCenterX, binCenterY);
      h[i].v=r.GetBinContent(binX, binY);
      h[i].e=r.GetBinError(binX, binY);
    }
  }
}


/// Convert BLOP 2D histogram to ROOT histogram.
void
b2r_convert(const hist& h, TH2D& r)
{
  const unsigned int nBinsX = h.bins(1)-2;
  const unsigned int nBinsY = h.bins(2)-2;
  r = TH2D(h.legend().c_str(),
           (h.legend()&";"&h.axistitle(1)&";"&h.axistitle(2)&";"&h.rangetitle()).c_str(),
           nBinsX,
           h.min(1)+(h.max(1)-h.min(1))/h.bins(1),
           h.max(1)-(h.max(1)-h.min(1))/h.bins(1),
           nBinsY,
           h.min(2)+(h.max(2)-h.min(2))/h.bins(2),
           h.max(2)-(h.max(2)-h.min(2))/h.bins(2));
  for ( unsigned int binX=0 ; binX<=nBinsX+1 ; ++binX )
  {
    for ( unsigned int binY=0 ; binY<=nBinsY+1 ; ++binY )
    {
      const double binCenterX = r.GetXaxis()->GetBinCenter(binX);
      const double binCenterY = r.GetYaxis()->GetBinCenter(binY);
      const unsigned int i = h.bin(binCenterX, binCenterY);
      r.SetBinContent(binX, binY, h[i].v);
      r.SetBinError(binX, binY, h[i].e);
    }
  }
}


/// Convert ROOT 3D histogram to BLOP histogram.
void
r2b_convert(const TH3D& r, hist& h)
{
  const unsigned int nBinsX = r.GetNbinsX();
  const unsigned int nBinsY = r.GetNbinsY();
  const unsigned int nBinsZ = r.GetNbinsZ();
  h.reinit(r.GetXaxis()->GetBinLowEdge(0),
           r.GetXaxis()->GetBinLowEdge(nBinsX+1)+r.GetXaxis()->GetBinWidth(nBinsX+1),
           nBinsX+2,
           r.GetXaxis()->GetTitle(),
           r.GetYaxis()->GetBinLowEdge(0),
           r.GetYaxis()->GetBinLowEdge(nBinsY+1)+r.GetYaxis()->GetBinWidth(nBinsY+1),
           nBinsY+2,
           r.GetYaxis()->GetTitle(),
           r.GetZaxis()->GetBinLowEdge(0),
           r.GetZaxis()->GetBinLowEdge(nBinsZ+1)+r.GetZaxis()->GetBinWidth(nBinsZ+1),
           nBinsZ+2,
           r.GetZaxis()->GetTitle(),
           "",
           r.GetTitle());
  for ( unsigned int binX=0 ; binX<=nBinsX+1 ; ++binX )
  {
    for ( unsigned int binY=0 ; binY<=nBinsY+1 ; ++binY )
    {
      for ( unsigned int binZ=0 ; binZ<=nBinsZ+1 ; ++binZ )
      {
        const double binCenterX = r.GetXaxis()->GetBinCenter(binX);
        const double binCenterY = r.GetYaxis()->GetBinCenter(binY);
        const double binCenterZ = r.GetZaxis()->GetBinCenter(binZ);
        const unsigned int i = h.bin(binCenterX, binCenterY, binCenterZ);
        h[i].v=r.GetBinContent(binX, binY, binZ);
        h[i].e=r.GetBinError(binX, binY, binZ);
      }
    }
  }
}


/// Convert BLOP 3D histogram to ROOT histogram.
void
b2r_convert(const hist& h, TH3D& r)
{
  const unsigned int nBinsX = h.bins(1)-2;
  const unsigned int nBinsY = h.bins(2)-2;
  const unsigned int nBinsZ = h.bins(3)-2;
  r = TH3D(h.legend().c_str(),
           (h.legend()&";"&h.axistitle(1)&";"&h.axistitle(2)&";"&h.axistitle(3)&";"&h.rangetitle()).c_str(),
           nBinsX,
           h.min(1)+(h.max(1)-h.min(1))/h.bins(1),
           h.max(1)-(h.max(1)-h.min(1))/h.bins(1),
           nBinsY,
           h.min(2)+(h.max(2)-h.min(2))/h.bins(2),
           h.max(2)-(h.max(2)-h.min(2))/h.bins(2),
           nBinsZ,
           h.min(3)+(h.max(3)-h.min(3))/h.bins(3),
           h.max(3)-(h.max(3)-h.min(3))/h.bins(3));
  for ( unsigned int binX=0 ; binX<=nBinsX+1 ; ++binX )
  {
    for ( unsigned int binY=0 ; binY<=nBinsY+1 ; ++binY )
    {
      for ( unsigned int binZ=0 ; binZ<=nBinsZ+1 ; ++binZ )
      {
        const double binCenterX = r.GetXaxis()->GetBinCenter(binX);
        const double binCenterY = r.GetYaxis()->GetBinCenter(binY);
        const double binCenterZ = r.GetZaxis()->GetBinCenter(binZ);
        const unsigned int i = h.bin(binCenterX, binCenterY, binCenterZ);
        r.SetBinContent(binX, binY, binZ, h[i].v);
        r.SetBinError(binX, binY, binZ, h[i].e);
      }
    }
  }
}


/// Read all ROOT histograms from a root file into BLOP histograms.
//  (Note: cannot read into ROOT histograms because ROOT is incompatible.
//   with itself: cannot append a ROOT histo to a vector/map.)
void
readHistos(const string& fileName, map<string, hist>& histos)
{
  TFile file(fileName.c_str());
  const TList& listOfKeys = *file.GetListOfKeys();
  TIter nextKey(&listOfKeys);
  TKey* key = 0;
  while ( ( key = (TKey*)nextKey() ) ) {
    const TObject* object = key->ReadObj();
    const std::string name = key->GetName();
    if ( object->IsA()->InheritsFrom(TH3::Class()) ) {
      TH3D* rootHisto = (TH3D*)object;
      hist& blopHisto = histos[name];
      r2b_convert(*rootHisto, blopHisto);
      cout << "Found TH3D: " << name << " in " << fileName << endl;
    } else if ( object->IsA()->InheritsFrom(TH2::Class()) ) {
      TH2D* rootHisto = (TH2D*)object;
      hist& blopHisto = histos[name];
      r2b_convert(*rootHisto, blopHisto);
      cout << "Found TH2D: " << name << " in " << fileName << endl;
    } else if ( object->IsA()->InheritsFrom(TH1::Class()) ) {
      TH1D* rootHisto = (TH1D*)object;
      hist& blopHisto = histos[name];
      r2b_convert(*rootHisto, blopHisto);
      cout << "Found TH1D: " << name << " in " << fileName << endl;
    } else {
      cerr << "Found non-TH1D,TH2D,TH3D object " << name << " in " << fileName << endl;
    }
  }
  file.Close();
}


}
